(declare (unit clucklib))

(include "logger.sci")

#>
#include <xmmsclient/xmmsclient.h>
#include <curses.h>
<#

; An error function for xmmsclient calls
(define (my-xmms-error-func error)
  (logger "In my-xmms-wait, found an error.\n")
  (popupLabel
    cdk_screen
    (make_strings
      (list
	(conc "Got an error from xmms2; exiting.  Error is:  " error)))
    1)
  (logger "Got error from xmms2: " (xmmsc_result_get_error call) "; exiting.\n")
  (exit 1))

; my-xmms-wait with a built-in error call
(define (my-xmms-wait call)
  (xmms-wait call my-xmms-error-func))

; xmms-wait when we don't care about the result
(define (my-xmms-wait-toss call)
  (xmmsc_result_unref (xmms-wait call my-xmms-error-func)))

; The most-standard highlight function: all entries in the current
; playlist get the good highlight
(define (highlight-current-playlist-good id info current-playlist-ids currently-playing-id)
  (logger "In highlight-current-playlist-good: " id ", " info nl)
  (cond
    [(member id current-playlist-ids)
     (conc good-highlight info)]
    [else info]))

; Given a list of media ids, refresh the selection list
; with that data.
;
; We also grab the list of songs in the current playlist, so we can
; highlight them.
;
; Arguments:
;
; 	raw-ids:	Media library ids of the songs to show
; 			information on.
; 			
; 	highlight:	A fuction of 4 arguments, used to select
; 			whether, and how, to highlight a row
; 			(actually can do arbitrary things to a row
; 			just before display).  See
; 			highlight-current-playlist-good for an
; 			example.
;
; 	no-sort?:	If #t, don't sort the results.
;
; 	extra-columns:	If not #f, should be a list of columns to
; 			add to the beginning of the columns of media
; 			information.
(define (show-media raw-ids highlight #!optional no-sort? extra-columns)
  ; If no media, reset to the main menu
  (cond
    [(eq? '() raw-ids)
     (popupLabel cdk_screen (make_strings (list "No music to display.")) 1)
     (show-menu)]
    [else
      (logger "show-media raw ids: " raw-ids nl)
      (logger "show-media optionsals: " no-sort? nl)
      (let* (
	     [ids (cond
		    [no-sort? raw-ids]
		    [else (media-sort raw-ids)])]
	     [null2 (logger "show-media ids: " ids nl)]
	     ; Get the media info, including the title row
	     [infos-pre (get-mediainfo-list ids music-columns)]
	     [null3 (logger "infos-pre: " infos-pre nl)]
	     [infos (cond
		      [extra-columns
			(map
			  cons
			  extra-columns infos-pre)]
		      [else infos-pre])]
	     [null3 (logger "infos: " infos nl)]
	     ; Get the columnar version of the media info
	     [pretty-infos (columnar-print maxwidth infos "")]
	     ; Get the full playlist
	     [current-ids (playlist-get-ids)]
	     [null2 (logger "In show-media, current ids: " current-ids nl)]
	     [real-id (playlist-current-id)]
	     ; Get the highlighting working
	     [pretty-infos2
	       (map
		 (lambda (id info)
		   ; Pass some extra values on to highlight
		   (highlight id info current-ids real-id))
		 ; Add an extra line to ids for the titles
		 (cons -1 ids)
		 pretty-infos)])
	(logger "ids: " (length ids) ", " ids nl)
	(logger "pi: " (length pretty-infos) ", " pretty-infos nl)
	(logger "pi2: " pretty-infos2 nl)

	(set-selection
	  (first pretty-infos2)
	  ids
	  (cdr pretty-infos2)))]))

; Use sort-columns to sort the given list of media ids
(define (media-sort raw-ids)
  (letrec
    ([sort-list
       (map
	 (lambda (x)
	   ; second is to drop the headers
	   (cons x (second (get-mediainfo-list (list x) sort-columns))))
	 raw-ids)]
     [media-less-real?
       (lambda (a b)
	 (cond
	   ; If either element is empty, it doesn't matter what we
	   ; return, so return #f
	   [(or (eq? '() a) (eq? '() b)) #f]
	   ; If the first elements are the same, recurse with the
	   ; remainder
	   [(string= (first a) (first b))
	    (media-less-real? (cdr a) (cdr b))]
	   ; First is less
	   ; Comparitor depends on whether both look like numbers
	   [(cond
	      [(and (string->number (first a)) (string->number (first b)))
	       (< (string->number (first a)) (string->number (first b)))]
	      [else (string< (first a) (first b))])
	    #t]
	   ; Second is less
	   [else #f]))]
     [media-less?
       (lambda (a b)
	 (logger "media-less?: " a ", " b nl)
	 ; Strip the first element, since that's always the id, and
	 ; pass on to the real worker.
	 (media-less-real? (cdr a) (cdr b)))])
    (map
      first
      (sort sort-list media-less?))))

; Uses get-mediainfo to get a whole list of ids and prepend a
; hard-coded title.
(define (get-mediainfo-list ids columns)
  (logger "In get-mediainfo-list, passed ids: " ids nl)
  (logger "In get-mediainfo-list, passed columns: " columns nl)
  (cons
    ; We append this here and then pull it out later so the
    ; header will have the same spacings as everything else.
    (map first columns)
    ; For each ID, get the values
    (map (lambda (id)
	   (cond
	     [id
	       ; For each ID + column, get the value
	       (get-mediainfo id (map second columns))]
	     [else (make-list (length (map second columns)) "")]))
	 ids)))

; Given a media id, returns artist, title, etc.
(define (get-mediainfo id keys)
  (logger "get-mediainfo: " id ", " keys nl)
  (define-external val c-string "N/A")
  (define intval (s32vector 999999))
  (define uintval (u32vector 999999))
  (define result (xmmsc_medialib_get_info conn id))
  (xmmsc_result_wait result)

  (let*
    ([out-result-raw
       (map
	 (lambda (key)
	   (logger "get-mediainfo key: " key nl)
	   ; Is this an error?
	   (cond [(= (xmmsc_result_iserror result) 1)
		  no-data]
		 ; No error found
		 [else
		   (let* ([type (xmmsc_result_get_dict_entry_type result key)]
			  [final-result
			    ; Do special processing for special cases.
			    (mediainfo-extras
			      key
			      (cond 
				; Pick amongst the various result types,
				; returning a string regardless.
				[(eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_NONE" int))
				 no-data]
				[(eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_UINT32" int))
				 (cond
				   ((= 0 (xmmsc_result_get_dict_entry_int result key uintval))
				    (quote no-data))
				   (else
				     (->string (u32vector-ref intval 0))))]
				[(eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_INT32" int))
				 (cond
				   ((= 0 (xmmsc_result_get_dict_entry_int result key intval))
				    (quote no-data))
				   (else
				     (->string (s32vector-ref intval 0))))]
				[(eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_STRING" int))
				 (cond
				   ((= 0 (xmmsc_result_get_dict_entry_string result key (location val)))
				    (quote no-data))
				   (else
				     val))]))])

		     ;(logger "get-mediainfo: " final-result nl)

		     (string-trim-both final-result))]))
	 keys)]
     [null (logger "In get-mediainfo: out-result-raw: " out-result-raw nl)]
     ; Always return a list of the length we were asked for
     [out-result
       (append
	 out-result-raw
	 (make-list (- (length keys) (length out-result-raw)) ""))]
     [null (logger "In get-mediainfo: out-result: " out-result nl)]
     )
    (xmmsc_result_unref result)

    out-result))

; Extra processing for special cases.
(define (mediainfo-extras key string)
  ;(logger "mediainfo-extras: " key nl)
  (cond
    [(and (string= key "duration") string (string->number string))
     (fmt #f
	  (join
	    dsp
	    (list
	      ; Minutes
	      (quotient (string->number string) 60000)
	      ":"
	      ; Seconds
	      (pad-char #\0 (pad/left 2 (dsp (remainder (quotient (string->number string) 1000) 60)))))))]
    [else string]))

; Callback for propdict_foreach
(define-external
  (propdict_foreach
    (c-string key)
    (int type)
    (c-pointer value)
    (c-string source)
    (c-pointer user_data))
  void
  (logger "In propdict_foreach.\n")
  (propdict-process key type value))

; The Scheme bit used by the callback for propdict_foreach
(define (propdict-process key type value)
  (logger "propdict-add: " key nl)
  (cond
    [(or
       ; We have no interest, ever, in dealing with properties that we
       ; have no idea how to display
       (eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_UINT32" int))
       (eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_INT32" int))
       (eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_STRING" int)))

     ; Pull various other values out before setting all-props
     (define-external val_ext c-pointer value)
     (cond
       [(string= key "artist")
	(global-set!
	  'all-artists
	  (lset-adjoin
	    string=
	    all-artists
	    ; Need to convert to a c-string
	    (string-trim-both (foreign-value "(char *)val_ext" c-string))))]

       ; Get tags; tags are defined as integer properties that we've
       ; only ever seen at 0 or 1
       [(and
	  (or
	    (eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_UINT32" int))
	    (eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_INT32" int)))
	  (or
	    (= 0 (foreign-value "(int *)val_ext" int))
	    (= 1 (foreign-value "(int *)val_ext" int)))
	  ; If we've seen this before (it's in all-props) then we
	  ; either have already added it to tags, or we must have seen
	  ; a non-tag-like value; either way, don't add it
	  (not (any (lambda (x) (string= key (first x))) all-props))
	  ; Don't add to all-tags props known not to be tag-like
	  (not (string= key "isvbr"))
	  (not (string= key "status")))
	(global-set! 'all-tags (lset-adjoin string= all-tags key))]

       ; Drop non-tag-like integers from all-tags
       [(and
	  (or
	    (eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_UINT32" int))
	    (eq? type (foreign-value "XMMSC_RESULT_VALUE_TYPE_INT32" int)))
	  (not (= 0 (foreign-value "(int *)val_ext" int)))
	  (not (= 1 (foreign-value "(int *)val_ext" int))))
	(logger "In propdict-process, " key " is not a tag." nl)
	(global-set! 'all-tags (delete key all-tags))])

     ; Now that we're done with tests that might rely on all-props
     ; not being set yet, set it.
     (global-set! 'all-props (lset-adjoin
			       (lambda (x y) (string= (first x) (first y)))
			       all-props
			       ; We need the type as well as the key,
			       ; unfortunately, for later use;
			       ; essentially, this makes all-props an
			       ; alist
			       (list
				 (string-trim-both key)
				 type)))]))

; This isn't actually useful unless organized by artist
; [(string= key "album")
;  (define-external val_ext c-pointer value)
;  (global-set! 'all-albums (lset-adjoin string= all-albums (foreign-value "(char *)val_ext" c-string)))]))

; Run through all the props for all the songs, collecting data.
;
; This actually gets a few more besides all-props, but I couldn't
; thing of a good name.  :)
(define (get-all-props)
  (logger "In get-all-props.\n")
  (global-set! 'all-artists '())
  (global-set! 'all-tags '())
  (global-set! 'all-props '())
  (let* ([null1 (logger "About to xmmsc_coll_universe.\n")]
	 ; Get a collection reference
	 [collref (xmmsc_coll_universe)]
	 ; Get the ids from the collection
	 [result (my-xmms-wait (xmmsc_coll_query_ids conn collref #f 0 0))]
	 [null1 (logger "Result: " result nl)]
	 [colls (result->int-list result)])
    ; Loop through all the songs
    (for-each
      (lambda (id)
	(let* ([result2 (my-xmms-wait (xmmsc_medialib_get_info conn id))])
	  ; Do a foreach over each prop of the song
	  (xmmsc_result_propdict_foreach
	    result2
	    (location propdict_foreach)
	    #f)
	  (xmmsc_result_unref result2)))
      colls)
    (xmmsc_result_unref result)))



; Return the length of this string with non-printing magic CDK bits
; removed.
(define (my-string-length str)
  (string-length (string-substitute "<[/A-Za-z0-9]+>" "" str)))

; columnar-print maxwidth items
;
; Takes a list of the form:
;
; (
;  ("ID" "Artist" ...)
;  ("1" "Sting" ....)
;  ...
; )
;
; and turns it into a list of the same length, with each element a
; columnated string.  Note that all cropping is done from the right,
; which is kind of stupid.
;
; BETTER ALG:
;
; Take a list of the item maxwidths, an list of the same length, but
; filled with zeros, and the line maxwidth. Div maxwidth by #items,
; round down.  Find all items that are less than the div value.
; Recurse with the lesser ones turned into zeros in the first list,
; their values put into the second list (maintaining position), and
; maxwidth = maxwidth - (+ all itemmaxwidths that we're moving from
; one list to the other). Bottom of recursion is when either the
; first list is all zeros or all non-zero items in the first list
; have width > maxwidth / #items; in the latter case, set each
; item's width to (maxwidth / #items) before moving to the second
; list and returning.  Result should be fair division of the total
; width.
(define (columnar-print maxwidth items prefix)
  (logger "In columnar-print: " items nl)
  (letrec
    ; Get the max lengths of the segments; return as a list of lists
    ; like: '((4 0) (17 0)...), for processing later.
    ([lengths (map
		(lambda (x) (list x 0))
		(fold
		  (lambda (item accum)
		    (cond
		      ; Bottom case
		      [(eq? #f accum)
		       (map my-string-length item)]
		      [else
			; For each item element, get the greater of
			; the item's length or the length we've
			; accumulated so far.
			(map
			  (lambda (x)
			    (max
			      (my-string-length (first x))
			      (second x)))
			  (zip item accum)) ]))
		  #f
		  items))]
     [null (logger "Lengths: " lengths "\n")]
     [fair-divide (lambda (mymaxw pairs)
		    (logger "In fair-divide: " mymaxw ", " pairs nl)
		    (let*
		      (; Get the pair bits seperately
		       [maxes (map first pairs)]
		       [fairs (map second pairs)]
		       [null (logger "maxes, fairs: " maxes ", " fairs nl)]
		       [nonzero? (lambda (y) (not (eq? 0 y)))]
		       [null (logger "non-zeroes: " (filter nonzero? maxes) "\n")]
		       ; Are they all zero?
		       [all-zero? (not (any nonzero? maxes))]
		       ; Divide the local maxwidth by the number of
		       ; remaining (i.e. non-zero) elements
		       [ceil (cond
			       ; Divide by 0 bad
			       [all-zero? 1]
			       [else (quotient
				       mymaxw
				       (max 1 (length (filter nonzero? maxes))))])]
		       [null (logger "ceil: " ceil nl)]
		       [null (logger "all zero: " all-zero? nl)]
		       ; Is every non-zero element bigger than the
		       ; ceiling?
		       [all-bigger? (every (lambda (y) (>= y ceil)) (filter nonzero? maxes))]
		       [null (logger "all bigger: " all-bigger? nl)]
		       ; Get all the first elements smaller than
		       ; ceil, to add them up
		       [smallers (filter (lambda (x) (< x ceil)) maxes)])
		      (cond
			; If all of the first elements are zero,
			; we're done; return the second elements.
			[all-zero? fairs]
			[else
			  ; Make a new set and recurse.
			  (fair-divide
			    (- mymaxw (apply + smallers))
			    (map
			      (lambda (maxitem fair)
				(logger "maxitem, fair: " maxitem ", " fair nl)
				(cond
				  ; If all remaining are bigger than
				  ; the ceiling, set them all to the
				  ; ceiling.
				  [(and all-bigger? (nonzero? maxitem))
				   (list 0 ceil)]
				  ; Else, move over those that are
				  ; less
				  [(and (nonzero? maxitem) (< maxitem ceil))
				   (list 0 maxitem)]
				  ; Else leave them alone.
				  [else (list maxitem fair)]))
			      maxes
			      fairs))])))]
     [fair-lengths
       ; Do fair division of the screen width, but don't use the full
       ; maxwidth: -3 for each boundary, and - the prefix length
       (fair-divide
	 (-
	   maxwidth
	   (* 3 (- (length (first items)) 1))
	   (string-length prefix)
	   4)
	 lengths)]
     [null (logger "Fair-lengths: " fair-lengths ", " (apply + fair-lengths) "\n")]
     [fmt-song (lambda (song)
		 (let*
		   ([song-fit (map
				(lambda (element len)
				  ;(logger "element len: " element ", " len nl)
				  (fmt #f (fit len element)))
				song
				fair-lengths)])
		   (logger "song-fit: " song-fit nl)
		   (fmt #f
			(fit
			  (- maxwidth 3)
			  (join
			    dsp
			    (list prefix
				  (join
				    dsp
				    song-fit
				    " | ")))))))]
     [pretty-songs (map fmt-song items)]
     )
    (logger "Pretty-songs: "
	    (join dsp (take pretty-songs (min 3 (length pretty-songs))) nl) "..." nl)
    (logger "Pretty-songs lens: "
	    (join dsp (map string-length (take pretty-songs (min 3 (length pretty-songs)))) nl) "..." nl)
    pretty-songs))

(define last_items #f)

; Display information to the selection widget
(define (set-selection title ids items)
  (logger "In set-selection: " (length items) ".\n")

  ; Store the item ids, for callbacks to use.
  (global-set! 'selection-items ids)
  (logger "Done set.\n")

  ; Spaces are for the choice string
  (logger "Setting title: " (conc "   " title) nl)
  (setCDKSelectionTitle selection-widget (conc "</B>" "   " title))

  ;(logger "Setting items: " items "\n")
  ;(logger "Old items: " last_items "\n")
  ;(cond [(not (eq? #f (global-ref 'last_items)))
  ;(logger "Freeing old items.\n")
  ;(CDKfreeStrings (global-ref 'last_items))])

  (cond
    ; If either are empty
    [(or
       (eq? '() ids)
       (eq? '() items))
     (global-set! 'selection-items (list #f))
     (setCDKSelectionItems selection-widget (make_strings (list "--- NO PLAYLISTS ---")) 1)]
    [else
      (define-external items_ext c-pointer (make_strings items))
      (setCDKSelectionItems selection-widget items_ext (length items))])
  ;(global-set! 'last_items items_ext))
  )

; Takes an xmms2 result that contains a string and extracts it
(define (result->string result)
  (xmms-result->string result my-xmms-error-func))

; Takes an xmms2 result that contains an int and extracts it
(define (result->int result)
  (xmms-result->int result my-xmms-error-func))

; Takes an xmms2 result that contains a list of ints and extracts them
(define (result->int-list result)
  (xmms-result->int-list result my-xmms-error-func))

; Takes an xmms2 result that contains a list of strings and extracts them
  (define (result->string-list result)
  (xmms-result->string-list result my-xmms-error-func))

; Takes a bunch of auxillary functions and walks through the list of
; currently selected choices on the given widget
;
; Arguments:
;
; 	object:  	The widget/object to handle; generally the
; 			argument to a callback.
;
; 	none-cont?:	Boolean: If nothing is chocies-selected, do
; 			we continue with the current cursor
; 			selection?
;
;	process-items:	Function of one argument; gets given the
;			entire list of items while it is still a
;			list of active choice numbers.
;
;	process-item:	Function of one argument; gets given a
;			single item from the results of
;			process-items.
;
(define (handle-choices object none-cont? process-items process-item)
  ; Why *this* run to logger needs the list, I have no idea.
  (logger (list "In handle-choices: "  object ", " none-cont? ", " process-items ", " process-item nl))
  (define-external obj2 c-pointer object)
  (let*
    ([choices-raw (get-selection-choices-list object)]
     [null (logger "In handle-choices choices-raw: " choices-raw nl)]
     [choices (cond
		; If no selections were made, either continue with
		; the current under-the-cursor selection, or not.
		[(null? choices-raw)
		 (cond
		   [none-cont? (list (getCDKSelectionCurrent object))]
		   [else #f])]
		; Else use the selections
		[else choices-raw])]
     [null (logger "In handle-choices choices: " choices nl)]
     ; Transform the list of choices per the user
     [processed-items (process-items choices)]
     [null (logger "In handle-choices processed-items: " processed-items nl)]
     )
    ; We might have set the choices to null above to exit.
    (cond
      [choices
	; For each selection, do something
	(for-each
	  ; Skip #f items entirely
	  (lambda (x)
	    (cond
	      [x (process-item x)]))
	  processed-items)]
      ; We have a #f choices; do nothing.
      [else #f]))

  ; Clear the selections now that they've been used.
  (clear-selection-choices-list object)
  (refresh-current))


;****************************
; find-source
;
; Horrific hack follows, but it's not my fault.
;
; In order to get the source of a particular property on a
; particular song, we have to iterate over *all* properties on that
; song.  Using a callback.  Which has to set a global to return any
; data.  GAaAahhH.
;
; This is as of 0.2 DrJekyll; may be better later.
;****************************

; The global we return
(define find-source-prop #f)
(define find-source-result #f)

; Callback for propdict_foreach
(define-external
  (propfind_foreach
    (c-string key)
    (int type)
    (c-pointer value)
    (c-string source)
    (c-pointer user_data))
  void
  (logger "In propfind_foreach.\n")
  (cond
    [(string= find-source-prop key)
     (set! find-source-result source)]))

; Run through all the props for the song, looking for the one that
; matches and returning the source.
(define (find-source id prop)
  (logger "In find-source: " id ", " prop nl)
  ; Set the  global.  *sigh*
  (set! find-source-prop prop)
  (set! find-source-result #f)
  (let* ([result (my-xmms-wait (xmmsc_medialib_get_info conn id))])
    ; Do a foreach over each prop of the song
    (xmmsc_result_propdict_foreach
      result
      (location propfind_foreach)
      #f)
    (xmmsc_result_unref result))
  (or
    find-source-result
    ; If none is set, set the default
    "client/cluck"))

; Our own getString call, that handles cursor issues.
(define (my-getString title label initial-string)
  (foreign-code "curs_set(1);")
  (let
    ([ret (getString cdk_screen title label initial-string)])
    (foreign-code "curs_set(0);")
    ret))
