(declare (unit directory))

#>
#include <xmmsclient/xmmsclient.h>
<#

(include "logger.sci")

(define directory-current #f)

(define current-entries #f)

; I don't know why, but putting this in directory-medialib-check,
; where it's actually used, seems to make us prone to segfaults.
(define-external temp_col c-pointer)

; Helper function: is this file in the media library?
(define (directory-medialib-check file)
  (let* (
	 [fixed-file (string-substitute*
		       file
		       ; Get our regexes escaped
		       (map
			 (lambda (entry)
			   (cons (regexp-escape (car entry)) (cdr entry)))
			 xmms-file-name-encoding)
		       #f)]
	 [query (conc "url:\"file://" fixed-file "\"")]
	 [null (logger "In directory-medialib-check query: " query nl)]
	 ; We need to not barf on a bad parse
	 [parse-ok? (eq? 1 (xmmsc_coll_parse query (location temp_col)))]
	 [null (logger "In directory-medialib-check parse-ok?: " parse-ok? nl)]
	 [result (and parse-ok? (my-xmms-wait (xmmsc_coll_query_ids conn temp_col #f 0 0)))]
	 [coll-ids (and parse-ok? (result->int-list result))]
	 )

    (logger "In directory-medialib-check media ids: " coll-ids nl)

    (xmmsc_result_unref result)
    (xmmsc_coll_unref temp_col)

    (cond
      ; The parse failed
      [(not coll-ids)
       (logger "In directory-medialib-check, failed\n")
       (list 'failed '())]
      ; If the id list is empty, item is not in the media library
      [(eq? coll-ids '())
       (logger "In directory-medialib-check, not found\n")
       (list 'not-found '())]
      ; We found something; should only ever return 1 item, but
      ; we're not actually going to worry about that.
      [(> (length coll-ids) 0)
       (logger "In directory-medialib-check, found\n")
       ; Return the ids and the fact that we found something
       (list 'found coll-ids)])))

(define (directory-highlight id info current-playlist-ids currently-playing-id)
  (logger "In directory-highlight: " id ", " info nl)
  (cond
    ; A directory == other1
    [(and (not id) (string-search "^[^|]*/" info))
     (logger "Looks like a directory.\n")
     (conc other1-highlight info)]
    ; Currently playing == other2
    [(= id currently-playing-id)
     (conc other2-highlight info)]
    ; In playlist == good
    [(member id current-playlist-ids)
     (conc good-highlight info)]
    ; Not in media == bad
    [(not id)
     (conc bad-highlight info)]
    [else info]))

; Show the current directory
(define (show-directory)
  (logger "In show-directory.\n")
  ; If a starting directory has not been chosen, ask for one
  (cond
    [(not starting-directory)
     (let get-dir
       ([dir-name (my-getString "Enter Starting Directory Name:" "Name: " "")])
       ; Is this a real dir?
       (cond
	 [(directory? dir-name)
	  ; Visit it to get the official name
	  (let*
	    ([oldpath (current-directory)]
	     [null (change-directory dir-name)]
	     [newpath (current-directory)]
	     [null (change-directory oldpath)])
	    (global-set! 'starting-directory newpath))
	  (set! directory-current starting-directory)]
	 [else
	   (popupLabel cdk_screen (make_strings (list "Directory not found.")) 1)
	   (get-dir (my-getString "Enter Starting Directory Name:" "Name: " ""))]))]
    [(not directory-current)
     (set! directory-current starting-directory)])
  (let*
    ; Get the directory entries, and add .. at the top.
    ([entries (sort (cons ".." (directory directory-current)) string<)]
     ; Append / to the directories
     [marked-entries
       (map
	 (lambda (entry)
	   (cond
	     ; If the entry is a directory, mark it as such.
	     [(directory? (conc directory-current "/" entry))
	      (conc entry "/")]
	     [else entry]))
	 entries)]
     ; Get the full playlist for comparison
     [current-ids (playlist-get-ids)]
     ; Get medialib ids for each song, or not
     [media-ids (map
		  (lambda (entry)
		    (cond
		      ; If the entry is not a regular file, we need no further processing
		      [(not (regular-file? (conc directory-current "/" entry)))
		       #f]
		      [else
			(let*
			  ([check (directory-medialib-check (conc directory-current "/" entry))])
			  (cond
			    ; Parse failed or item not found
			    [(or
			       (eq? (first check) 'failed)
			       (eq? (first check) 'not-found))
			     #f]
			    [else (first (second check))]))]))
		  entries)])
    (logger "In show-directory, entries: " entries ", " marked-entries nl)
    (global-set! 'status-extra (conc "Directory: " directory-current))
    ; Set the regular set-selection to the media id
    (show-media media-ids directory-highlight #t (cons "File/Dir" marked-entries))
    ; And store the directory entries in our own special place
    (set! current-entries entries)
    )
  (show-directory-help)
  )

; Add a file to the media library
;
; We assume the caller knows whether the file is in the library or
; not, although a double-add does no damage
(define (directory-media-add file after)
  (my-xmms-wait (xmmsc_medialib_add_entry conn (conc "file://" file)))
  (let*
    ([media-check (directory-medialib-check file)])
    (logger "In directory-media-add for " file ", media-check " media-check nl)
    (cond
      ; If either form of failure occured, die.
      [(or
	 (eq? (first media-check) 'not-found)
	 (eq? (first media-check) 'failed))
       (popupLabel cdk_screen (make_strings (list (conc "Media import of " file " failed, sorry."))) 1)]
      ; Otherwise, do whatever the after-add function says
      [else
	(after media-check)])))

; Add the list of files to the media library
(define (directory-media-add-object object)
  (handle-choices
    object
    #t
    directory-expand-items
    ; Just run directory-media-add with a no-op "after"
    (lambda (x)
      (directory-media-add (conc directory-current "/" x) (lambda (x) #t))))
  (show-directory))

; Dig into the next directory
(define (directory-enter object)
  (let*
    ([selection-id (getCDKSelectionCurrent object)]
     [entering (list-ref current-entries selection-id)]
     [new-file (conc directory-current "/" entering)]
     [media-check (directory-medialib-check new-file)])
    (logger "In directory-enter, media-check: " media-check nl)
    (cond
      ; Backing out one
      [(or
	 (string= entering "../")
	 (string= entering ".."))
       (directory-up)]
      ; Entering a normal directory
      [(directory? new-file)
       (set!
	 directory-current
	 (conc directory-current "/" entering)) 
       (show-directory)]
      ; Not in the media library: add it
      [(or
	 (eq? (first media-check) 'not-found)
	 (eq? (first media-check) 'failed))
       (directory-media-add new-file)]
      ; "Entering" a file in the media library: play the song
      [(file-exists? new-file)
       (playlist-add object)
       (setCDKSelectionCurrent selection-widget selection-id)
       (playlist-play object)
       (show-directory)])))

; Turn the list of choices into a list of directory entries,
; including recursive descent of same.
(define (directory-expand-items items-raw #!optional no-convert)
  (let walk-items
    ([items
       (cond
	 [no-convert
	   items-raw]
	 [else
	   (map
	     (lambda (x)
	       (list-ref current-entries x))
	     items-raw)])])
    (logger "In directory-expand-items, items: " items nl)
    (cond
      ; We're done
      [(eq? '() items) '()]
      ; The item is a directory: grab all the entries, and run
      ; walk-items on them, recursively.
      [(directory? (conc directory-current "/" (first items)))
       (logger "In directory-expand-items, item is a directory.\n")
       (flatten
	 (walk-items
	   (map
	     (lambda (x)
	       (conc (first items) "/" x))
	     (directory (conc directory-current "/" (first items)))))
	 (walk-items (cdr items)))]
      ; The item is a file: just cons the string
      [(file-exists? (conc directory-current "/" (first items)))
       (logger "In directory-expand-items, item is a file.\n")
       (cons (first items) (walk-items (cdr items)))])))

; Create a new directory
(define (directory-new)
  (create-directory
    (conc
      directory-current
      "/"
      (my-getString "Enter New Directory Name:" "Name: " "")))
  (show-directory))

; Rename an item, including directories.
;
; The tricky part here is updating the url fields in the media
; library
(define (directory-rename object)
  (logger "In directory-rename.\n")
  (let*
    ([renaming (list-ref current-entries (getCDKSelectionCurrent object))]
     [new-name (my-getString "Enter New Name:" "Name: " renaming)]
     [old-path (conc directory-current "/" renaming)]
     [new-path (conc directory-current "/" new-name)]
     )
    (logger "In directory-rename: renaming " renaming " to " new-name nl)
    ; We defer all the real work to directory-media-rename because
    ; copy/paste operations are just special cases of renaming, so
    ; might as well do all the work there.
    (directory-media-rename old-path new-path new-name)))

; Do the actual renaming of media
(define (directory-media-rename old-path new-path new-name)
  (logger "In directory-media-rename: renaming " old-path " to " new-path nl)
  (cond
    ; Easy case: regular file.
    [(and
       (file-exists? old-path)
       (not (directory? old-path)))
     (rename-file old-path new-path)
     ; We'll likely have barfed if it didn't work, but check
     ; anyways.
     (cond
       [(and
	  (file-exists? new-path)
	  (not (directory? new-path)))
	; Update the media library
	(directory-media-url-change old-path new-path)])]
    ; Harder case: directory file.
    [(directory? old-path)
     ; SOLUTION SKETCH: move the directory, then (if that works)
     ; descend to each non-directory item underneath and do the
     ; media library update.
     (rename-file old-path new-path)
     (cond
       [(directory? new-path)
	; Given the list of things under new-path, run
	; directory-media-url-change on s/new-path/old-path/ and
	; new-path.
	(map
	  (lambda (path)
	    (logger "In directory-media-rename: directory type, path: " path ", " old-path ", " new-path nl)
	    ; We're not interested in anything but the last segment
	    ; of path, which is always going to be something like
	    ; "dir/foo", where dir is the last element in both
	    ; old-path and new-path.
	    (directory-media-url-change
	      (conc old-path "/" (string-substitute ".*/" "" path))
	      (conc new-path "/" (string-substitute ".*/" "" path))))
	  (directory-expand-items (list new-name) #t))
	])
     #f]
    ))

; Change the URL tag for the media entry that matches old-path to be
; new-path instead.  file:// is added.  Do *NOT* encode the path!
(define (directory-media-url-change old-path new-path)
  (logger "In directory-media-url-change: renaming " old-path " to " new-path nl)
  (let*
    ([old-info (directory-medialib-check old-path)])
    (logger "In directory-media-url-change: old-info: " old-info nl)
    (cond
      ; There's no good way to deal with it not having been found;
      ; we have to assume it was never in the media library and move
      ; on.
      [(eq? (first old-info) 'found)
       ; Set the url property, with "server" as the source.
       (my-xmms-wait
	 (xmmsc_medialib_move_entry
	   conn				; connection
	   (first (second old-info))	; song id
	   (conc
	     "file://"
	     new-path)))]))		; url
  (show-directory))

(define cut-items '())
; Cut the given item(s): grab the list of things selected and stick
; them in a list for later pasting.
(define (directory-cut object)
  (logger "In directory-cut\n")
  (handle-choices
    object
    #t
    ; Turn the list of choices into a list of directory entries
    (lambda (items)
      (map
	(lambda (x)
	  (list-ref current-entries x))
	items))
    ; For each directory entry, add it with the full path to
    ; cut-items
    (lambda (entry)
      (logger "In directory-cut, adding " entry nl)
      (set! cut-items (lset-adjoin string= cut-items (conc directory-current "/" entry)))
      (logger "In directory-cut, cut-items: " cut-items nl))))

; Paste all the things that we've "cut" into the current directory.
;
; In other words, a bunch of moves.
(define (directory-paste)
  (logger "In directory-paste\n")
  (map
    (lambda (item)
      (logger "In directory-paste, item: " item nl)
      (directory-media-rename
	item
	; Strip the directory, and stick the current one on
	(conc
	  directory-current
	  "/"
	  (string-substitute ".*/" "" item))
	; For the short name, just strip the dir
	(string-substitute ".*/" "" item)))
    cut-items))

; Back out of the current directory, or back to the main menu
(define (directory-up)
  (logger "In directory-up, cwd: " directory-current ", " starting-directory nl)
  (cond
    ; If at the top, back to the main menu
    [(string= directory-current starting-directory)
     (show-menu)]
    [else
      (set!
	directory-current
	(string-substitute "/[^/][^/]*/?$" "" directory-current))
      (show-directory)]))

; Delete (actually move) all selected songs
(define (directory-delete-song object)
  (logger "In directory-delete-song\n")
  (handle-choices
    object
    #t
    ; Turn the list of choices into a list of directory entries
    (lambda (items)
      (map
	(lambda (x)
	  (list-ref current-entries x))
	items))
    ; For each directory entry, move it and then delete all the
    ; relevant songs
    (lambda (entry)
      (let* ([items (directory-expand-items (list entry) #t)])
	(logger "In directory-cut, moving " entry ", " items nl)
	(rename-file (conc directory-current "/" entry) (conc "/tmp/" entry))
	(popupLabel cdk_screen (make_strings (list "Moved to /tmp/; just in case.")) 1)
	; OK, we've moved the files; now we delete the medialib
	; entries; have to descend dirs for that
	(for-each
	  (lambda (x)
	    (let* ([check (directory-medialib-check (conc directory-current "/" x))])
	      (cond
		[(eq? (first check) 'found)
		 (my-xmms-wait (xmmsc_medialib_remove_entry conn (first (second check))))])))
	  items)))))

; Add all songs down the tree to the playlist
(define (directory-playlist-add object)
  (logger "In directory-playlist-add\n")
  (handle-choices
    object
    #t
    directory-expand-items
    (lambda (x)
      (let* ([check (directory-medialib-check (conc directory-current "/" x))])
	(cond
	  [(eq? (first check) 'found)
	   (my-xmms-wait-toss (xmmsc_playlist_add_id conn (foreign-value "XMMS_ACTIVE_PLAYLIST" c-string) (first (second check))))])))))
