(declare (unit status))

(include "logger.sci")

(use fmt)
(use srfi-18)

(define status-current-id #f)

#>
#include <curses.h>
#include <cdk.h>
#include <xmmsclient/xmmsclient.h>
<#

; Go directly to curses for the status refresh.
#>!
int wrefresh(WINDOW *win);
<#

; Handle the callback from the loop we hacked in to cdk's
; selection.c
(define-external
  (handle_waiting (c-pointer object)
		  (c-pointer clientData))
  void
  (logger "In handle_waiting.\n")
  (status-update))

; Set up the alarm handler
(define (status-setup)
  (logger "In status-setup.\n")
  ; Run the update right away.
  (status-update)
  (setCDKObjectWaiting selection-widget (location handle_waiting) #f))

; Store the current song; if it changes, refresh all
(define status-current-song -1)

; The actual alarm handler: determines, and displays, current status
; information.
(define (status-update)
  (logger "In status-update\n")
  (thread-sleep! 0.4)
  ; Sync the collections about one time in N; easier than counting.
  (cond [(= 0 (random 50))
	 (my-xmms-wait (xmmsc_coll_sync conn))
	 (logger "In status-update, done sync.\n")])
  (cond
    [(not (= (playlist-current-id) status-current-song))
     (refresh-current)
     (set! status-current-song (playlist-current-id))])


  ; Make the cursor go the hell away
  (foreign-code "curs_set( 0 );" )
  (let*
    ([real-id (playlist-current-id)]
     [raw-title (first (get-mediainfo real-id (list "title")))]
     [song-artist (first (get-mediainfo real-id (list "artist")))]
     [song-name (fmt #f
		     (cond
		       [(string= no-data raw-title)
			(string-substitute
			  ".*/" ""
			  (first (get-mediainfo real-id (list "url"))))]
		       [else
			 (first (get-mediainfo real-id (list "title")))]))]
     [song-time 
       (fmt #f
	    (join
	      dsp
	      (list
		; Pretending this is a duration
		(mediainfo-extras "duration" (->string (result->int (xmmsc_playback_playtime conn))))
		" out of "
		; Get the actual duration
		(first (get-mediainfo real-id (list "duration"))))))]
     [playback-status (status-get-playback-status)]
     [general-status
       (fmt #f
	    (join
	      dsp
	      (remove
		; Drop emtpy strings
		(lambda (x) (string= x ""))
		(list
		  ; Extra bits set by other sections of code
		  (cond [status-extra status-extra] [else ""])
		  ; Basic play state
		  (list-ref
		    (list "Stopped" "Playing" "Paused")
		    playback-status)
		  ; Auto-writeback
		  (cond [prop-writeback "Auto-FileWrite"]
			[else ""])
		  ; Auto-Rating
		  (cond [auto-rating "Auto-Rating"]
			[else ""])
		  ; Auto-Skip
		  (cond [auto-skip "Auto-Skip"]
			[else ""])
		  ; Looping
		  (cond [(string= "1" (result->string (xmmsc_configval_get conn "playlist.repeat_all")))
			 "Loop All"]
			[else ""])
		  (cond [(string= "1" (result->string (xmmsc_configval_get conn "playlist.repeat_one")))
			 "Loop One"]
			[else ""])
		  ))
	      ", "))]
     [help-status
       (cond
	 [(not little-help)
	  "?: Small Help - H: Big Help"]
	 [else ""])]
     ; Setup to get the file and directory names
     [fixed-file (string-substitute*
		   (string-substitute
		     "^file://" ""
		     (first (get-mediainfo real-id (list "url"))))
		   ; Get our regexes escaped
		   (map
		     (lambda (entry)
		       (cons (regexp-escape (cdr entry)) (car entry)))
		     xmms-file-name-encoding)
		   #f)]
     ; Get the directory name
     [directory-name (string-substitute "/[^/]*$" "" fixed-file)]
     ; Get the file name
     [file-name (string-substitute ".*/" "" fixed-file)]
     )

    ; Set the current song directory
    (global-set! 'current-song-directory directory-name)

    ; Now that we have the id, we can try to auto-skip
    (ratings-auto-skip real-id)

    (global-set! 'status-current-id real-id)
    (logger "status-alarm song-name: " song-name nl)

    (define-external status_widget c-pointer status-widget)
    (define-external status_window c-pointer (foreign-value "((CDKLABEL *)status_widget)->win" c-pointer))

    ; Artist line
    (status-write-line 1 "Artist: " song-artist "Directory: " directory-name)

    ; Title line
    (status-write-line 2 "Title: " song-name "File: " file-name)

    ; Time line
    (status-write-line 3 "Time: " song-time)

    ; Status line
    (status-write-line 4 "Status: " general-status)

    ; Special upper help line
    (writeChar
      status_window
      2
      0
      help-status
      (foreign-value "HORIZONTAL" int)
      0
      (string-length help-status))

    (wrefresh status_window)))

; Get the status integer from the server
(define (status-get-playback-status)
  (result->int (my-xmms-wait (xmmsc_playback_status conn))))

(define (status-write-line line-num title data #!optional right-title right-data)
  (logger "In status-write-line: "  line-num ", " title ", " data ", " right-title ", " right-data nl)
  (writeCharAttrib
    status_window
    2
    line-num
    title
    (foreign-value "A_BOLD" int)
    (foreign-value "HORIZONTAL" int)
    0
    (string-length title))
  (writeChar
    status_window
    (+ 2 (string-length title))
    line-num
    (fmt #f (trim (- maxwidth (string-length title)) data))
    (foreign-value "HORIZONTAL" int)
    0
    (string-length data))

  ; If we still have room, add the right-justified bits the horror
  ; that follows is due to the fact that if we want to have the
  ; title highlighted we need to handle the right-justification
  ; ourselves
  (cond
    [(and
       right-title
       (< (+ 20 (string-length title) (string-length data)) maxwidth))
     (logger "In status-write-line: doing right-just bits.\n")
     (let*
       ([after-data (+ 2 (string-length title) (string-length data))]
	[remaining (- maxwidth after-data)]
	[gap (max 1 (- remaining
		       (+
			 (string-length right-title)
			 (string-length right-data))))]
	)
       (logger "In status, right-just stats: " after-data ", " remaining ", " gap ", " (string-length right-data) nl)
       (writeChar
	 status_window
	 after-data
	 line-num
	 (fmt #f (fit gap ""))
	 (foreign-value "HORIZONTAL" int)
	 0
	 gap)
       (writeCharAttrib
	 status_window
	 (+ gap after-data)
	 line-num
	 right-title
	 (foreign-value "A_BOLD" int)
	 (foreign-value "HORIZONTAL" int)
	 0
	 (string-length right-title))
       (writeChar
	 status_window
	 (+ (string-length right-title) gap after-data)
	 line-num
	 (fmt #f (fit (- remaining (string-length right-title) gap) right-data))
	 (foreign-value "HORIZONTAL" int)
	 0
	 (- remaining (string-length right-title) gap))
       )]
    ; Else fill out the rest of the line
    [else
      (let*
	([after-data (+ 2 (string-length title) (string-length data))]
	 [rest (max 0 (- maxwidth after-data))])
	(writeChar
	  status_window
	  after-data
	  line-num
	  (fmt #f (fit rest ""))
	  (foreign-value "HORIZONTAL" int)
	  0
	  rest))
      ]))
